(in-package :araneida)

(defgeneric start-listening (listener &key &allow-other-keys))
(defgeneric stop-listening (listener &key abort &allow-other-keys))
(defvar *break-on-handler-errors* nil)

(defmacro with-accept-flets (&body body)
  `(labels ((do-it (listener s)
	      (let ((r (read-request-from-stream listener s)))
		(handler-case
		    (handle-request-using-listener
		     listener (http-listener-handler listener) r)
		  (http-error (c) 
		    (request-send-error r (http-error-code c) 
					(http-error-message c))))))
	    (accept (listener)
	      (let ((socket (socket-accept (http-listener-socket listener))))
		(socket-make-stream socket :element-type 'base-char
				    :name "socket"
				    :input t :output t :buffering :full))))
     (with-simple-restart
	 (abort-response "Abort this response and answer another request")
       ;; expectation is that socket-accept will not block, because we 
       ;; are invoked when select() says something is ready.  we really
       ;; ought to set the master socket non-blocking to be sure.
       (let ((*debugger-hook* #'handler-debugger-hook))
	 ,@body))))

(defgeneric handle-request-using-listener (http-listener handler request))

(defmethod handle-request-using-listener ((l http-listener) handler request)
  (handle-request (http-listener-handler l) request))

(defun handler-debugger-hook (condition old-hook)
  (declare (ignore old-hook))
  (unless *break-on-handler-errors*
    (typecase condition
      (sb-bsd-sockets::bad-file-descriptor-error 
       (format *trace-output*
	       "accept: bad file descriptor error (socket shut down?)~%~A~%"
	       condition))
      (sb-bsd-sockets::socket-error
       (format *trace-output* "socket error ~A~%" condition))
      (sb-ext:timeout
       (format *trace-output* "timed out processing request: ~A~%" condition))
      (t
       (format *trace-output* "Caught error: ~A~%" condition)
       (sb-debug:backtrace 15 *trace-output*)))
    (invoke-restart 'abort-response)))





